@@ -20,6 +20,7 @@ import os
 import threading
 import time
 import uuid
+from urllib.parse import urlparse
 from xml.etree import ElementTree as ET
 
 import lasso
@@ -27,15 +28,13 @@ import requests
 import requests.exceptions
 from atomicwrites import atomic_write
 from django.contrib import auth, messages
+from django.contrib.auth import get_user_model
 from django.contrib.auth.models import Group
 from django.core.exceptions import FieldDoesNotExist, PermissionDenied
 from django.core.files.storage import default_storage
-from django.utils import six
-from django.utils.encoding import force_text
-from django.utils.six.moves.urllib.parse import urlparse
-from django.utils.translation import ugettext as _
+from django.utils.translation import gettext as _
 
-from . import app_settings, models, utils
+from . import app_settings, models, models_utils, utils
 
 User = auth.get_user_model()
 
@@ -56,6 +55,8 @@ def display_truncated_list(l, max_length=10):
 
 
 class DefaultAdapter:
+    user_class = get_user_model()
+
     def __init__(self, request=None):
         self.request = request
 
@@ -267,6 +268,11 @@ class DefaultAdapter:
         if required_classref:
             given_classref = saml_attributes['authn_context_class_ref']
             if given_classref is None or given_classref not in required_classref:
+                logger.info(
+                    'mellon: refused login because of authn_classref mismatch (%r vs %s)',
+                    given_classref,
+                    required_classref,
+                )
                 raise PermissionDenied
         return True
 
@@ -274,8 +280,8 @@ class DefaultAdapter:
         realm = utils.get_setting(idp, 'REALM')
         username_template = utils.get_setting(idp, 'USERNAME_TEMPLATE')
         try:
-            username = force_text(username_template).format(realm=realm, attributes=saml_attributes, idp=idp)[
-                :30
+            username = username_template.format(realm=realm, attributes=saml_attributes, idp=idp)[
+                : self.user_class._meta.get_field('username').max_length
             ]
         except ValueError:
             logger.error('mellon: invalid username template %r', username_template)
@@ -310,6 +316,7 @@ class DefaultAdapter:
                             transient_federation_attribute,
                         )
                         return None
+                saml_attributes['transient_name_id_content'] = name_id
             else:
                 if self.request:
                     messages.warning(
@@ -322,14 +329,25 @@ class DefaultAdapter:
                 return None
         else:
             name_id = saml_attributes['name_id_content']
-        issuer = saml_attributes['issuer']
+        entity_id = saml_attributes['issuer']
         try:
+            to_update = {
+                'nid_format': saml_attributes['name_id_format'],
+                'nid_name_qualifier': saml_attributes.get('name_id_name_qualifier'),
+                'nid_sp_name_qualifier': saml_attributes.get('name_id_sp_name_qualifier'),
+                'nid_sp_provided_id': saml_attributes.get('name_id_sp_provided_id'),
+            }
             saml_identifier = models.UserSAMLIdentifier.objects.select_related('user').get(
-                name_id=name_id, issuer=issuer
+                name_id=name_id, issuer=models_utils.get_issuer(entity_id)
             )
+            # nid_* attributes are new, we must update them if they are not initialized, eventually
+            for key in to_update:
+                if getattr(saml_identifier, key) != to_update[key]:
+                    models.UserSAMLIdentifier.objects.filter(pk=saml_identifier.pk).update(**to_update)
+                    break
             user = saml_identifier.user
             user.saml_identifier = saml_identifier
-            logger.info('mellon: looked up user %s with name_id %s from issuer %s', user, name_id, issuer)
+            logger.info('mellon: looked up user %s with name_id %s from issuer %s', user, name_id, entity_id)
             return user
         except models.UserSAMLIdentifier.DoesNotExist:
             pass
@@ -344,10 +362,10 @@ class DefaultAdapter:
             created = True
             user = self.create_user(User)
 
-        nameid_user = self._link_user(idp, saml_attributes, issuer, name_id, user)
+        nameid_user = self._link_user(idp, saml_attributes, user)
         if user != nameid_user:
             logger.info(
-                'mellon: looked up user %s with name_id %s from issuer %s', nameid_user, name_id, issuer
+                'mellon: looked up user %s with name_id %s from issuer %s', nameid_user, name_id, entity_id
             )
             if created:
                 user.delete()
@@ -360,7 +378,7 @@ class DefaultAdapter:
                 user.delete()
                 return None
             logger.info(
-                'mellon: created new user %s with name_id %s from issuer %s', nameid_user, name_id, issuer
+                'mellon: created new user %s with name_id %s from issuer %s', nameid_user, name_id, entity_id
             )
         return nameid_user
 
@@ -452,10 +470,29 @@ class DefaultAdapter:
             )
         return None
 
-    def _link_user(self, idp, saml_attributes, issuer, name_id, user):
+    def _link_user(self, idp, saml_attributes, user):
+        name_id_content = saml_attributes['name_id_content']
+        if saml_attributes['name_id_format'] == lasso.SAML2_NAME_IDENTIFIER_FORMAT_TRANSIENT:
+            name_id_content = saml_attributes['transient_name_id_content']
+        to_update = {
+            'nid_format': saml_attributes['name_id_format'],
+            'nid_name_qualifier': saml_attributes.get('name_id_name_qualifier'),
+            'nid_sp_name_qualifier': saml_attributes.get('name_id_sp_name_qualifier'),
+            'nid_sp_provided_id': saml_attributes.get('name_id_sp_provided_id'),
+        }
         saml_id, created = models.UserSAMLIdentifier.objects.get_or_create(
-            name_id=name_id, issuer=issuer, defaults={'user': user}
+            name_id=name_id_content,
+            issuer=models_utils.get_issuer(saml_attributes['issuer']),
+            defaults={
+                'user': user,
+                **to_update,
+            },
         )
+        # nid_* attributes are new, we must update them eventually
+        for key in to_update:
+            if getattr(saml_id, key) != to_update[key]:
+                models.UserSAMLIdentifier.objects.filter(pk=saml_id.pk).update(**to_update)
+                break
         if created:
             user.saml_identifier = saml_id
             return user
@@ -474,7 +511,7 @@ class DefaultAdapter:
         attribute_set = False
         for field, tpl in attribute_mapping.items():
             try:
-                value = force_text(tpl).format(realm=realm, attributes=saml_attributes, idp=idp)
+                value = tpl.format(realm=realm, attributes=saml_attributes, idp=idp)
             except ValueError:
                 logger.warning('mellon: invalid attribute mapping template %r', tpl)
             except (AttributeError, KeyError, IndexError, ValueError) as e:
