@@ -1,8 +1,5 @@
-# -*- coding: utf-8 -*-
-from __future__ import unicode_literals
-
-from django.db import models, migrations
 from django.conf import settings
+from django.db import migrations, models
 
 
 class Migration(migrations.Migration):
@@ -15,11 +12,22 @@ class Migration(migrations.Migration):
         migrations.CreateModel(
             name='UserSAMLIdentifier',
             fields=[
-                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
+                (
+                    'id',
+                    models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True),
+                ),
                 ('issuer', models.TextField(verbose_name='Issuer')),
                 ('name_id', models.TextField(verbose_name='SAML identifier')),
                 ('created', models.DateTimeField(auto_now_add=True, verbose_name='created')),
-                ('user', models.ForeignKey(related_name='saml_identifiers', verbose_name='user', to=settings.AUTH_USER_MODEL, on_delete=models.CASCADE)),
+                (
+                    'user',
+                    models.ForeignKey(
+                        related_name='saml_identifiers',
+                        verbose_name='user',
+                        to=settings.AUTH_USER_MODEL,
+                        on_delete=models.CASCADE,
+                    ),
+                ),
             ],
             options={
                 'verbose_name': 'user SAML identifier',
@@ -29,6 +37,6 @@ class Migration(migrations.Migration):
         ),
         migrations.AlterUniqueTogether(
             name='usersamlidentifier',
-            unique_together=set([('issuer', 'name_id')]),
+            unique_together={('issuer', 'name_id')},
         ),
     ]
