@@ -18,6 +18,7 @@ import datetime
 import importlib
 import logging
 from functools import wraps
+from urllib.parse import urlparse
 from xml.parsers import expat
 
 import isodate
@@ -26,8 +27,7 @@ from django.conf import settings
 from django.contrib import auth
 from django.template.loader import render_to_string
 from django.urls import reverse
-from django.utils.encoding import force_text
-from django.utils.six.moves.urllib.parse import urlparse
+from django.utils.encoding import force_str
 from django.utils.timezone import get_default_timezone, is_aware, make_aware, make_naive, now
 
 from . import app_settings
@@ -35,6 +35,10 @@ from . import app_settings
 logger = logging.getLogger(__name__)
 
 
+class CreateServerError(Exception):
+    pass
+
+
 def create_metadata(request):
     entity_id = reverse('mellon_metadata')
     login_url = reverse(app_settings.LOGIN_URL)
@@ -84,6 +88,8 @@ def create_server(request):
         server = lasso.Server.newFromBuffers(
             metadata, private_key_content=private_key, private_key_password=private_key_password
         )
+        if not server:
+            raise CreateServerError
         if app_settings.SIGNATURE_METHOD:
             symbol_name = 'SIGNATURE_METHOD_' + app_settings.SIGNATURE_METHOD.replace('-', '_').upper()
             if hasattr(lasso, symbol_name):
@@ -205,26 +211,8 @@ def get_setting(idp, name, default=None):
     return idp.get(name) or getattr(app_settings, name, default)
 
 
-def make_session_dump(lasso_name_id, indexes):
-    session_infos = []
-    name_id = force_text(lasso_name_id.content)
-    name_id_format = force_text(lasso_name_id.format)
-    name_qualifier = lasso_name_id.nameQualifier and force_text(lasso_name_id.nameQualifier)
-    sp_name_qualifier = lasso_name_id.spNameQualifier and force_text(lasso_name_id.spNameQualifier)
-    for index in indexes:
-        issuer = index.saml_identifier.issuer
-        session_infos.append(
-            {
-                'entity_id': issuer,
-                'session_index': index.session_index,
-                'name_id_content': name_id,
-                'name_id_format': name_id_format,
-                'name_id_name_qualifier': name_qualifier,
-                'name_id_sp_name_qualifier': sp_name_qualifier,
-            }
-        )
-    session_dump = render_to_string('mellon/session_dump.xml', {'session_infos': session_infos})
-    return session_dump
+def make_session_dump(session_indexes):
+    return render_to_string('mellon/session_dump.xml', {'session_indexes': session_indexes})
 
 
 def create_logout(request):
@@ -239,18 +227,67 @@ def is_nonnull(s):
     return '\x00' not in s
 
 
+PROTOCOLS_TO_PORT = {
+    'http': '80',
+    'https': '443',
+}
+
+
+def netloc_to_host_port(netloc):
+    if not netloc:
+        return None, None
+    splitted = netloc.split(':', 1)
+    if len(splitted) > 1:
+        return splitted[0], splitted[1]
+    return splitted[0], None
+
+
+def same_domain(domain1, domain2):
+    if domain1 == domain2:
+        return True
+
+    if not domain1 or not domain2:
+        return False
+
+    if domain2.startswith('.'):
+        # p1 is a sub-domain or the base domain
+        if domain1.endswith(domain2) or domain1 == domain2[1:]:
+            return True
+    return False
+
+
 def same_origin(url1, url2):
-    """
-    Checks if two URLs are 'same-origin'
+    """Checks if both URL use the same domain. It understands domain patterns on url2, i.e. .example.com
+    matches www.example.com.
+
+    If not scheme is given in url2, scheme compare is skipped.
+    If not scheme and not port are given, port compare is skipped.
+    The last two rules allow authorizing complete domains easily.
     """
     p1, p2 = urlparse(url1), urlparse(url2)
-    if url1.startswith('/') or url2.startswith('/'):
+    p1_host, p1_port = netloc_to_host_port(p1.netloc)
+    p2_host, p2_port = netloc_to_host_port(p2.netloc)
+
+    # url2 is relative, always same domain
+    if url2.startswith('/') and not url2.startswith('//'):
         return True
-    try:
-        return (p1.scheme, p1.hostname, p1.port) == (p2.scheme, p2.hostname, p2.port)
-    except ValueError:
+
+    if p2.scheme and p1.scheme != p2.scheme:
         return False
 
+    if not same_domain(p1_host, p2_host):
+        return False
+
+    try:
+        if (p2_port or (p1_port and p2.scheme)) and (
+            (p1_port or PROTOCOLS_TO_PORT[p1.scheme]) != (p2_port or PROTOCOLS_TO_PORT[p2.scheme])
+        ):
+            return False
+    except (ValueError, KeyError):
+        return False
+
+    return True
+
 
 def get_status_codes_and_message(profile):
     assert profile, 'missing lasso.Profile'
