@@ -28,20 +28,21 @@ import requests
 from django.conf import settings
 from django.contrib import auth
 from django.contrib.auth import REDIRECT_FIELD_NAME, get_user_model
+from django.core import signing
 from django.db import transaction
 from django.http import Http404, HttpResponse, HttpResponseForbidden, HttpResponseRedirect
 from django.shortcuts import render, resolve_url
 from django.urls import reverse
-from django.utils import six
-from django.utils.encoding import force_str, force_text
+from django.utils.encoding import force_str
 from django.utils.http import urlencode
-from django.utils.translation import ugettext as _
+from django.utils.timezone import now
+from django.utils.translation import gettext as _
 from django.views.decorators.csrf import csrf_exempt
 from django.views.generic import View
 from django.views.generic.base import RedirectView
 from requests.exceptions import RequestException
 
-from . import app_settings, models, utils
+from . import app_settings, models, models_utils, utils
 
 RETRY_LOGIN_COOKIE = 'MELLON_RETRY_LOGIN'
 
@@ -85,7 +86,7 @@ def check_next_url(request, next_url):
     except UnicodeError:
         log.warning('next parameter ignored, as is\'s not an ASCII string')
         return
-    if not utils.same_origin(next_url, request.build_absolute_uri()):
+    if not utils.same_origin(request.build_absolute_uri(), next_url):
         log.warning('next parameter ignored as it is not of the same origin')
         return
     return next_url
@@ -133,6 +134,45 @@ class ProfileMixin:
             args.append(idp_message)
         self.log.warning(*args)
 
+    def dispatch(self, request, *args, **kwargs):
+        try:
+            return super().dispatch(request, *args, **kwargs)
+        except utils.CreateServerError:
+            return self.failure(
+                request,
+                reason=_(
+                    'Unable to initialize a SAML server object, the private key '
+                    'is maybe invalid or unreadable, please check its access '
+                    'rights and content.'
+                ),
+            )
+
+    def failure(self, request, reason='', status_codes=()):
+        '''show error message to user after a login failure'''
+        login = self.profile
+        idp = utils.get_idp(login and login.remoteProviderId)
+        if not idp and login:
+            self.log.warning('entity id %r is unknown', login.remoteProviderId)
+            return HttpResponseBadRequest('entity id %r is unknown' % login.remoteProviderId)
+        error_url = utils.get_setting(idp, 'ERROR_URL')
+        error_redirect_after_timeout = utils.get_setting(idp, 'ERROR_REDIRECT_AFTER_TIMEOUT')
+        if error_url:
+            error_url = resolve_url(error_url)
+        next_url = error_url or self.get_next_url(default=resolve_url(settings.LOGIN_REDIRECT_URL))
+        return self.render(
+            request,
+            'mellon/authentication_failed.html',
+            {
+                'debug': settings.DEBUG,
+                'reason': reason,
+                'status_codes': status_codes,
+                'issuer': login and login.remoteProviderId,
+                'next_url': next_url,
+                'relaystate': login and login.msgRelayState,
+                'error_redirect_after_timeout': error_redirect_after_timeout,
+            },
+        )
+
 
 class LoginView(ProfileMixin, LogMixin, View):
     def dispatch(self, request, *args, **kwargs):
@@ -193,38 +233,14 @@ class LoginView(ProfileMixin, LogMixin, View):
         ):
             self.show_message_status_is_not_success(login, 'SAML authentication failed')
         except lasso.Error as e:
+            if self.debug_login:
+                return self.render_debug_template(request, login)
             return HttpResponseBadRequest('error processing the authentication response: %r' % e)
         else:
             if 'RelayState' in request.POST and utils.is_nonnull(request.POST['RelayState']):
                 login.msgRelayState = request.POST['RelayState']
             return self.sso_success(request, login)
-        return self.sso_failure(request, reason=idp_message, status_codes=status_codes)
-
-    def sso_failure(self, request, reason='', status_codes=()):
-        '''show error message to user after a login failure'''
-        login = self.profile
-        idp = utils.get_idp(login.remoteProviderId)
-        if not idp:
-            self.log.warning('entity id %r is unknown', login.remoteProviderId)
-            return HttpResponseBadRequest('entity id %r is unknown' % login.remoteProviderId)
-        error_url = utils.get_setting(idp, 'ERROR_URL')
-        error_redirect_after_timeout = utils.get_setting(idp, 'ERROR_REDIRECT_AFTER_TIMEOUT')
-        if error_url:
-            error_url = resolve_url(error_url)
-        next_url = error_url or self.get_next_url(default=resolve_url(settings.LOGIN_REDIRECT_URL))
-        return self.render(
-            request,
-            'mellon/authentication_failed.html',
-            {
-                'debug': settings.DEBUG,
-                'reason': reason,
-                'status_codes': status_codes,
-                'issuer': login.remoteProviderId,
-                'next_url': next_url,
-                'relaystate': login.msgRelayState,
-                'error_redirect_after_timeout': error_redirect_after_timeout,
-            },
-        )
+        return self.failure(request, reason=idp_message, status_codes=status_codes)
 
     def get_attribute_value(self, attribute, attribute_value):
         # check attribute_value contains only text
@@ -244,7 +260,7 @@ class LoginView(ProfileMixin, LogMixin, View):
                     content = self.get_attribute_value(at, attribute_value)
                     if content is not None:
                         values.append(content)
-        attributes['issuer'] = login.remoteProviderId
+        entity_id = attributes['issuer'] = login.remoteProviderId
         in_response_to = login.response.inResponseTo
         if in_response_to:
             attributes['nonce'] = request.session.get('mellon-nonce-%s' % in_response_to)
@@ -252,14 +268,19 @@ class LoginView(ProfileMixin, LogMixin, View):
 
         if login.nameIdentifier:
             name_id = login.nameIdentifier
-            name_id_format = force_text(name_id.format or lasso.SAML2_NAME_IDENTIFIER_FORMAT_UNSPECIFIED)
+            name_id_format = force_str(name_id.format or lasso.SAML2_NAME_IDENTIFIER_FORMAT_UNSPECIFIED)
             attributes.update(
-                {'name_id_content': lasso_decode(name_id.content), 'name_id_format': name_id_format}
+                {
+                    'name_id_content': lasso_decode(name_id.content),
+                    'name_id_format': name_id_format,
+                }
             )
             if name_id.nameQualifier:
-                attributes['name_id_name_qualifier'] = force_text(name_id.nameQualifier)
+                attributes['name_id_name_qualifier'] = force_str(name_id.nameQualifier)
             if name_id.spNameQualifier:
-                attributes['name_id_sp_name_qualifier'] = force_text(name_id.spNameQualifier)
+                attributes['name_id_sp_name_qualifier'] = force_str(name_id.spNameQualifier)
+            if name_id.spProvidedId:
+                attributes['name_id_provided_id'] = force_str(name_id.spProvidedId)
         authn_statement = login.assertion.authnStatement[0]
         if authn_statement.authnInstant:
             attributes['authn_instant'] = utils.iso8601_to_datetime(authn_statement.authnInstant)
@@ -277,10 +298,14 @@ class LoginView(ProfileMixin, LogMixin, View):
         self.log.debug('trying to authenticate with attributes %r', attributes)
         response = self.authenticate(request, login, attributes)
         response.delete_cookie(RETRY_LOGIN_COOKIE)
+        if self.debug_login:
+            return self.render_debug_template(request, login, attributes)
         return response
 
     def authenticate(self, request, login, attributes):
-        user = auth.authenticate(request=request, saml_attributes=attributes)
+        user = auth.authenticate(
+            request=request, issuer=models_utils.get_issuer(attributes['issuer']), saml_attributes=attributes
+        )
         next_url = self.get_next_url(default=resolve_url(settings.LOGIN_REDIRECT_URL))
         if user is not None:
             if user.is_active:
@@ -295,14 +320,9 @@ class LoginView(ProfileMixin, LogMixin, View):
         else:
             self.log.warning('no user found for NameID %r', attributes['name_id_content'])
             return self.render(request, 'mellon/user_not_found.html', {'saml_attributes': attributes})
-        request.session['lasso_session_dump'] = login.session.dump()
+        return HttpResponseRedirect(next_url)
 
-        if self.debug_login:
-            return self.render_debug_template(request, login, attributes)
-        else:
-            return HttpResponseRedirect(next_url)
-
-    def render_debug_template(self, request, login, attributes):
+    def render_debug_template(self, request, login, attributes=None):
         request.session['mellon_debug_login'] = False
         context = {
             'logs': self.stream.getvalue(),
@@ -311,6 +331,9 @@ class LoginView(ProfileMixin, LogMixin, View):
             'response_dump': login.response and login.response.debug(4),
             'assertion_dump': login.assertion and login.assertion.debug(4),
         }
+        self.log.info('mellon: debug login attributes %s', attributes)
+        self.log.info('mellon: debug login assertion_dump %s', context['assertion_dump'])
+        self.log.info('mellon: debug login response_dump %s', context['response_dump'])
         return self.render(request, 'mellon/debug_login.html', context)
 
     def login(self, user, attributes):
@@ -319,6 +342,13 @@ class LoginView(ProfileMixin, LogMixin, View):
             return
 
         utils.login(self.request, user)
+        if (
+            app_settings.OPENED_SESSION_COOKIE_NAME
+            and app_settings.OPENED_SESSION_COOKIE_NAME in self.request.COOKIES
+        ):
+            self.request.session['mellon_opened_session_cookie'] = self.request.COOKIES[
+                app_settings.OPENED_SESSION_COOKIE_NAME
+            ]
         session_index = attributes['session_index']
         if session_index:
             if not self.request.session.session_key:
@@ -327,6 +357,10 @@ class LoginView(ProfileMixin, LogMixin, View):
                 saml_identifier=user.saml_identifier,
                 session_key=self.request.session.session_key,
                 session_index=session_index,
+                # keep transient nameid to be able to produce logout requests
+                transient_name_id=attributes['name_id_content']
+                if attributes['name_id_format'] == lasso.SAML2_NAME_IDENTIFIER_FORMAT_TRANSIENT
+                else None,
             )
         self.log.info('user %s (NameID is %r) logged in using SAML', user, attributes['name_id_content'])
         self.request.session['mellon_session'] = utils.flatten_datetime(attributes)
@@ -339,7 +373,7 @@ class LoginView(ProfileMixin, LogMixin, View):
         Use a cookie to prevent looping forever.
         """
         if RETRY_LOGIN_COOKIE in self.request.COOKIES:
-            response = self.sso_failure(
+            response = self.failure(
                 self.request, reason=_('There were too many redirections with the identity provider.')
             )
             response.delete_cookie(RETRY_LOGIN_COOKIE)
@@ -391,7 +425,7 @@ class LoginView(ProfileMixin, LogMixin, View):
             )
         except RequestException as e:
             self.log.warning('unable to reach %r: %s', login.msgUrl, e)
-            return self.sso_failure(
+            return self.failure(
                 request,
                 reason=_('IdP is temporarily down, please try again ' 'later.'),
                 status_codes=status_codes,
@@ -402,7 +436,7 @@ class LoginView(ProfileMixin, LogMixin, View):
                 result.status_code,
                 result.content,
             )
-            return self.sso_failure(request, reason=idp_message, status_codes=status_codes)
+            return self.failure(request, reason=idp_message, status_codes=status_codes)
 
         self.log.info('Got SAML Artifact Response', extra={'saml_response': result.content})
         result.encoding = utils.get_xml_encoding(result.content)
@@ -446,7 +480,7 @@ class LoginView(ProfileMixin, LogMixin, View):
             return HttpResponseBadRequest('error processing the authentication response: %r' % e)
         else:
             return self.sso_success(request, login)
-        return self.sso_failure(request, reason=idp_message, status_codes=status_codes)
+        return self.failure(request, reason=idp_message, status_codes=status_codes)
 
     def request_discovery_service(self, request, is_passive=False):
         return_url = request.build_absolute_uri()
@@ -587,18 +621,20 @@ login = transaction.non_atomic_requests(csrf_exempt(LoginView.as_view()))
 
 
 class LogoutView(ProfileMixin, LogMixin, View):
-    def get(self, request, *args, **kwargs):
-        if 'SAMLRequest' in request.GET:
+    def get(self, request, *args, logout_next_url='/', **kwargs):
+        if 'token' in request.GET:
+            return self.sp_logout_token(request, token=request.GET['token'], logout_next_url=logout_next_url)
+        elif 'SAMLRequest' in request.GET:
             return self.idp_logout(request, request.META['QUERY_STRING'], 'redirect')
         elif 'SAMLResponse' in request.GET:
-            return self.sp_logout_response(request)
+            return self.sp_logout_response(request, logout_next_url=logout_next_url)
         else:
-            return self.sp_logout_request(request)
+            return self.sp_logout_request(request, logout_next_url=logout_next_url)
 
     def post(self, request, *args, **kwargs):
         return self.idp_logout(request, force_str(request.body), 'soap')
 
-    def logout(self, request, issuer, saml_user, session_indexes, indexes, mode):
+    def logout(self, request, saml_user, session_indexes, indexes, mode):
         session_keys = set(indexes.values_list('session_key', flat=True))
         indexes.delete()
 
@@ -647,14 +683,15 @@ class LogoutView(ProfileMixin, LogMixin, View):
         except lasso.Error as e:
             return HttpResponseBadRequest('error processing logout request: %r' % e)
 
-        issuer = force_text(logout.remoteProviderId)
-        session_indexes = {force_text(sessionIndex) for sessionIndex in logout.request.sessionIndexes}
+        entity_id = force_str(logout.remoteProviderId)
+        session_indexes = {force_str(sessionIndex) for sessionIndex in logout.request.sessionIndexes}
 
         saml_identifier = (
             models.UserSAMLIdentifier.objects.filter(
-                name_id=force_text(logout.nameIdentifier.content), issuer=issuer
+                name_id=force_str(logout.nameIdentifier.content),
+                issuer=models_utils.get_issuer(entity_id),
             )
-            .select_related('user')
+            .select_related('user', 'issuer')
             .first()
         )
 
@@ -667,7 +704,7 @@ class LogoutView(ProfileMixin, LogMixin, View):
                 indexes = indexes.filter(session_index__in=session_indexes)
 
             # lasso has too much state :/
-            logout.setSessionFromDump(utils.make_session_dump(logout.nameIdentifier, indexes))
+            logout.setSessionFromDump(utils.make_session_dump(indexes))
 
             try:
                 logout.validateRequest()
@@ -680,7 +717,6 @@ class LogoutView(ProfileMixin, LogMixin, View):
                     self.log.info('full logout requested, no sessionIndexes')
                 self.logout(
                     request,
-                    issuer=issuer,
                     saml_user=name_id_user,
                     session_indexes=session_indexes,
                     indexes=indexes,
@@ -692,26 +728,41 @@ class LogoutView(ProfileMixin, LogMixin, View):
         except lasso.Error as e:
             return HttpResponseBadRequest('error processing logout request: %r' % e)
         if logout.msgBody:
-            return HttpResponse(force_text(logout.msgBody), content_type='text/xml')
+            return HttpResponse(force_str(logout.msgBody), content_type='text/xml')
         else:
             return HttpResponseRedirect(logout.msgUrl)
 
-    def sp_logout_request(self, request):
+    def next_url_cookie_name(self, relaystate):
+        return f'MellonNextURL-{relaystate}'
+
+    def sp_logout_request(self, request, logout_next_url=None):
         '''Launch a logout request to the identity provider'''
-        next_url = request.GET.get(REDIRECT_FIELD_NAME)
-        referer = request.META.get('HTTP_REFERER')
-        if not referer or utils.same_origin(referer, request.build_absolute_uri()):
+        referer = request.headers.get('Referer')
+        field_next_url = request.GET.get(REDIRECT_FIELD_NAME)
+        next_url = None
+        if field_next_url and utils.same_origin(request.build_absolute_uri(), field_next_url):
+            next_url = field_next_url
+        next_url = next_url or logout_next_url
+        if not referer or utils.same_origin(request.build_absolute_uri(), referer):
             if hasattr(request, 'user') and request.user.is_authenticated:
                 logout = None
                 try:
                     issuer = request.session.get('mellon_session', {}).get('issuer')
                     if issuer:
                         self.profile = logout = utils.create_logout(request)
+                        self.get_relay_state(create=True)
                         try:
-                            if 'lasso_session_dump' in request.session:
-                                logout.setSessionFromDump(request.session['lasso_session_dump'])
-                            else:
+                            session_indexes = models.SessionIndex.objects.filter(
+                                saml_identifier__user=request.user,
+                                saml_identifier__issuer__entity_id=issuer,
+                                session_key=request.session.session_key,
+                            )
+                            if not session_indexes:
                                 self.log.error('unable to find lasso session dump')
+                            else:
+                                session_dump = utils.make_session_dump(session_indexes)
+                                logout.setSessionFromDump(session_dump)
+                            session_indexes.update(logout_timestamp=now())
                             logout.initRequest(issuer, lasso.HTTP_METHOD_REDIRECT)
                             logout.buildRequestMsg()
                         except lasso.Error as e:
@@ -728,18 +779,21 @@ class LogoutView(ProfileMixin, LogMixin, View):
                     self.log.info('user logged out, SLO request sent to IdP')
             else:
                 # anonymous user: if next_url is None redirect to referer
-                return HttpResponseRedirect(next_url or referer)
+                return HttpResponseRedirect(next_url)
         else:
             self.log.warning('logout refused referer %r is not of the same origin', referer)
         return HttpResponseRedirect(next_url)
 
-    def sp_logout_response(self, request):
+    def sp_logout_response(self, request, logout_next_url='/'):
         '''Launch a logout request to the identity provider'''
         self.profile = logout = utils.create_logout(request)
+        logout.msgRelayState = request.GET.get('RelayState')
+        cookie_name = self.next_url_cookie_name(logout.msgRelayState)
+        cookie_next_url = request.COOKIES.get(cookie_name)
+        next_url = self.get_next_url() or cookie_next_url or logout_next_url
         # the user shouldn't be logged anymore at this point but it may happen
         # that a concurrent SSO happened in the meantime, so we do another
         # logout to make sure.
-        auth.logout(request)
         try:
             logout.processResponseMsg(request.META['QUERY_STRING'])
         except lasso.ProfileStatusNotSuccessError:
@@ -748,10 +802,69 @@ class LogoutView(ProfileMixin, LogMixin, View):
             self.log.warning('partial logout')
         except lasso.Error as e:
             self.log.warning('unable to process a logout response: %s', e)
-            return HttpResponseRedirect(resolve_url(settings.LOGIN_REDIRECT_URL))
-        next_url = self.get_next_url(default=resolve_url(settings.LOGIN_REDIRECT_URL))
+        response = HttpResponseRedirect(next_url)
+        if cookie_name in request.COOKIES:
+            response.delete_cookie(cookie_name)
+        models.SessionIndex.cleanup(chunk_size=100)
+        return response
+
+    TOKEN_SALT = 'mellon-logout-token'
+
+    def sp_logout_token(self, request, token, logout_next_url):
+        token_content = signing.loads(token, salt=self.TOKEN_SALT)
+        next_url = token_content['next_url'] or logout_next_url
+        session_index_pk = token_content['session_index_pk']
+        session_index = models.SessionIndex.objects.filter(pk=session_index_pk).first()
+        session_indexes = models.SessionIndex.objects.filter(
+            saml_identifier=session_index.saml_identifier, session_key=session_index.session_key
+        )
+        if session_indexes:
+            session_dump = utils.make_session_dump(session_indexes)
+            logout = utils.create_logout(request)
+            logout.msgRelayState = str(uuid.uuid4())
+            try:
+                logout.setSessionFromDump(session_dump)
+                logout.initRequest(
+                    session_indexes[0].saml_identifier.issuer.entity_id, lasso.HTTP_METHOD_REDIRECT
+                )
+                logout.buildRequestMsg()
+            except lasso.Error as e:
+                self.log.error('unable to initiate a logout request %r', e)
+                return HttpResponseRedirect(next_url)
+            except Exception:
+                self.log.exception('unable to initiate a logout request')
+                return HttpResponseRedirect(next_url)
+            else:
+                self.log.debug('sending LogoutRequest %r to URL %r', logout.request.dump(), logout.msgUrl)
+                response = HttpResponseRedirect(logout.msgUrl)
+                response.set_cookie(
+                    self.next_url_cookie_name(logout.msgRelayState),
+                    value=next_url,
+                    max_age=600,
+                    samesite='Lax',
+                )
+                return response
         return HttpResponseRedirect(next_url)
 
+    @classmethod
+    def make_logout_token_url(cls, request, next_url=None):
+        issuer = request.session.get('mellon_session', {}).get('issuer')
+        if not issuer:
+            return None
+        session_indexes = models.SessionIndex.objects.filter(
+            saml_identifier__user=request.user, saml_identifier__issuer__entity_id=issuer
+        ).order_by('-id')
+        if not session_indexes:
+            return None
+
+        token_content = {
+            'next_url': next_url,
+            'session_index_pk': session_indexes[0].pk,
+        }
+        token = signing.dumps(token_content, salt=cls.TOKEN_SALT)
+        session_indexes.update(logout_timestamp=now())
+        return reverse('mellon_logout') + '?' + urlencode({'token': token})
+
 
 logout = csrf_exempt(LogoutView.as_view())
 
