@@ -68,9 +68,9 @@ def jane(db):
 def test_format_username(settings, idp, saml_attributes):
     adapter = DefaultAdapter()
     assert adapter.format_username(idp, {}) is None
-    assert adapter.format_username(idp, saml_attributes) == ('x' * 32 + '@saml')[:30]
+    assert adapter.format_username(idp, saml_attributes) == ('x' * 32 + '@saml')
     settings.MELLON_USERNAME_TEMPLATE = '{attributes[name_id_content]}'
-    assert adapter.format_username(idp, saml_attributes) == ('x' * 32)[:30]
+    assert adapter.format_username(idp, saml_attributes) == ('x' * 32)
     settings.MELLON_USERNAME_TEMPLATE = '{attributes[username][0]}'
     assert adapter.format_username(idp, saml_attributes) == 'foobar'
 
@@ -92,10 +92,12 @@ def test_lookup_user(settings, idp, saml_attributes):
     assert User.objects.count() == 0
 
 
-def test_lookup_user_transaction(transactional_db, concurrency, idp, saml_attributes):
+def test_lookup_user_transaction(transactional_db, concurrency, idp, saml_attributes, settings):
     adapter = DefaultAdapter()
     p = ThreadPool(concurrency)
 
+    settings.MELLON_IDENTITY_PROVIDERS = [idp]
+
     if connection.vendor == 'postgresql':
         with connection.cursor() as c:
             c.execute('SHOW max_connections')
@@ -127,7 +129,7 @@ def test_provision_user_attributes(settings, django_user_model, idp, saml_attrib
         'last_name': '{attributes[last_name][0]}',
     }
     user = SAMLBackend().authenticate(saml_attributes=saml_attributes)
-    assert user.username == 'x' * 30
+    assert user.username == 'xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx@saml'
     assert user.first_name == 'Foo'
     assert user.last_name == 'Bar'
     assert user.email == 'test@example.net'
@@ -194,6 +196,8 @@ def test_provision_absent_attribute(settings, django_user_model, idp, saml_attri
 
 
 def test_provision_long_attribute(settings, django_user_model, idp, saml_attributes, caplog):
+    from django import VERSION
+
     settings.MELLON_IDENTITY_PROVIDERS = [idp]
     settings.MELLON_ATTRIBUTE_MAPPING = {
         'email': '{attributes[email][0]}',
@@ -203,11 +207,16 @@ def test_provision_long_attribute(settings, django_user_model, idp, saml_attribu
     local_saml_attributes = saml_attributes.copy()
     local_saml_attributes['first_name'] = [('y' * 32)]
     user = SAMLBackend().authenticate(saml_attributes=local_saml_attributes)
-    assert user.first_name == 'y' * 30
     assert len(caplog.records) == 4
     assert 'created new user' in caplog.text
     assert 'set field first_name' in caplog.text
-    assert 'to value %r ' % ('y' * 30) in caplog.text
+    if VERSION[0] <= 2:
+        assert user.first_name == 'y' * 30
+        assert 'to value %r ' % ('y' * 30) in caplog.text
+    else:
+        # django users' first name attribute longer from django3 onwards
+        assert user.first_name == 'y' * 32
+        assert 'to value %r ' % ('y' * 32) in caplog.text
     assert 'set field last_name' in caplog.text
     assert 'set field email' in caplog.text
 
