@@ -16,23 +16,25 @@
 
 import base64
 import datetime
+import logging
 import re
+import urllib.parse as urlparse
 import xml.etree.ElementTree as ET
 import zlib
 from html import unescape
+from unittest import mock
 
 import lasso
 import pytest
 from django.contrib.auth.models import User
 from django.contrib.sessions.models import Session
 from django.urls import reverse
-from django.utils import six
 from django.utils.encoding import force_str
-from django.utils.six.moves.urllib import parse as urlparse
 from httmock import HTTMock, all_requests
 from httmock import response as mock_response
 from pytest import fixture
 
+from mellon import models
 from mellon.utils import create_metadata
 from mellon.views import lasso_decode
 
@@ -82,7 +84,7 @@ class MockIdp:
     session_dump = None
     identity_dump = None
 
-    def __init__(self, idp_metadata, private_key, sp_metadata):
+    def __init__(self, idp_metadata, private_key, sp_metadata, name_id=None):
         self.server = server = lasso.Server.newFromBuffers(idp_metadata, private_key)
         self.server.signatureMethod = lasso.SIGNATURE_METHOD_RSA_SHA256
         server.addProviderFromBuffer(lasso.PROVIDER_ROLE_SP, sp_metadata)
@@ -90,7 +92,7 @@ class MockIdp:
     def reset_session_dump(self):
         self.session_dump = None
 
-    def process_authn_request_redirect(self, url, auth_result=True, consent=True, msg=None):
+    def process_authn_request_redirect(self, url, auth_result=True, consent=True, msg=None, name_id=None):
         login = self.login = lasso.Login(self.server)
         if self.identity_dump:
             login.setIdentityFromDump(self.identity_dump)
@@ -121,6 +123,8 @@ class MockIdp:
                 datetime.datetime.now().isoformat(),
                 datetime.datetime.now().isoformat(),
             )
+            for key in name_id or {}:
+                setattr(login.assertion.subject.nameID, key, name_id[key])
 
             def add_attribute(name, *values, **kwargs):
                 fmt = kwargs.get('fmt', lasso.SAML2_ATTRIBUTE_NAME_FORMAT_BASIC)
@@ -211,7 +215,16 @@ class MockIdp:
         if body:
             logout.processResponseMsg(force_str(body))
         else:
-            logout.processResponseMsg(force_str(url.split('?', 1)[-1]))
+            logout.processResponseMsg(url.split('?', 1)[-1])
+
+    def process_logout_request_redirect(self, url):
+        logout = lasso.Logout(self.server)
+        logout.setIdentityFromDump(self.identity_dump)
+        logout.setSessionFromDump(self.session_dump)
+        logout.processRequestMsg(url.split('?', 1)[1])
+        logout.validateRequest()
+        logout.buildResponseMsg()
+        return logout.msgUrl
 
     def mock_artifact_resolver(self):
         @all_requests
@@ -239,9 +252,50 @@ def test_sso_slo(db, app, idp, caplog, sp_settings):
     assert urlparse.urlparse(response['Location']).path == '/whatever/'
     response = app.get(reverse('mellon_logout'), extra_environ={'HTTP_REFERER': '/some/path'})
     assert urlparse.urlparse(response['Location']).path == '/singleLogout'
+    url = idp.process_logout_request_redirect(response.location)
+    response = app.get(url)
+    assert response.location == '/'
+
     # again, user is already logged out
     response = app.get(reverse('mellon_logout'), extra_environ={'HTTP_REFERER': '/some/path'})
-    assert urlparse.urlparse(response['Location']).path == '/some/path'
+    assert urlparse.urlparse(response['Location']).path == '/'
+
+
+def test_sso_slo_next(db, app, idp, caplog, sp_settings):
+    response = app.get(reverse('mellon_login'))
+    url, body, relay_state = idp.process_authn_request_redirect(response['Location'])
+    response = app.post(reverse('mellon_login'), params={'SAMLResponse': body, 'RelayState': relay_state})
+    response = app.get(
+        reverse('mellon_logout') + '?next=/some/path/', extra_environ={'HTTP_REFERER': '/other/path'}
+    )
+    assert urlparse.urlparse(response['Location']).path == '/singleLogout'
+    url = idp.process_logout_request_redirect(response.location)
+    response = app.get(url)
+    assert response.location == '/some/path/'
+
+
+def test_sso_slo_default_next_url(db, app, idp, caplog, sp_settings, rf, django_user_model):
+    from mellon.views import logout
+
+    response = app.get(reverse('mellon_login'))
+    url, body, relay_state = idp.process_authn_request_redirect(response['Location'])
+    response = app.post(reverse('mellon_login'), params={'SAMLResponse': body, 'RelayState': relay_state})
+
+    request = rf.get('/logout/')
+    request.session = app.session
+    request.user = django_user_model.objects.get()
+    response = logout(request, logout_next_url='/other/path/')
+    assert list(request.session.values()) == ['/other/path/']
+
+    response = app.get(reverse('mellon_login'))
+    url, body, relay_state = idp.process_authn_request_redirect(response['Location'])
+    response = app.post(reverse('mellon_login'), params={'SAMLResponse': body, 'RelayState': relay_state})
+
+    request = rf.get('/logout/?next=/some/path/')
+    request.session = app.session
+    request.user = django_user_model.objects.get()
+    response = logout(request, logout_next_url='/other/path/')
+    assert list(request.session.values()) == ['/some/path/']
 
 
 def test_sso_idp_slo(db, app, idp, caplog, sp_settings):
@@ -323,7 +377,7 @@ def test_sso_idp_slo_soap(db, app, idp, caplog, sp_settings):
     app.cookiejar.clear()
 
     url, body, relay_state = idp.init_slo(method=lasso.HTTP_METHOD_SOAP)
-    response = app.post(url, params=body, headers={'Content-Type': force_str('text/xml')})
+    response = app.post(url, params=body, headers={'Content-Type': 'text/xml'})
     assert Session.objects.count() == 1
     idp.check_slo_return(body=response.content)
 
@@ -401,7 +455,7 @@ def test_sso_idp_slo_full_soap(db, app, idp, caplog, sp_settings):
     # idp logout
     app.cookiejar.clear()
     url, body, relay_state = idp.init_slo(method=lasso.HTTP_METHOD_SOAP, full=True)
-    response = app.post(url, params=body, headers={'Content-Type': force_str('text/xml')})
+    response = app.post(url, params=body, headers={'Content-Type': 'text/xml'})
     assert Session.objects.count() == 0
     idp.check_slo_return(body=response.content)
 
@@ -653,13 +707,14 @@ def test_passive_auth_middleware_ok(db, app, idp, caplog, settings):
     settings.MELLON_OPENED_SESSION_COOKIE_NAME = 'IDP_SESSION'
     assert 'MELLON_PASSIVE_TRIED' not in app.cookies
     # webtest-lint is against unicode
-    app.set_cookie('IDP_SESSION', '1')
-    response = app.get('/', headers={'Accept': force_str('text/html')}, status=302)
+    app.set_cookie('IDP_SESSION', '1234')
+    response = app.get('/', headers={'Accept': 'text/html'}, status=302)
     assert urlparse.urlparse(response.location).path == '/login/'
     assert urlparse.parse_qs(urlparse.urlparse(response.location).query, keep_blank_values=True) == {
         'next': ['http://testserver/'],
         'passive': [''],
     }
+    assert app.cookies['MELLON_PASSIVE_TRIED'] == '1234'
 
     # simulate closing of session at IdP
     app.cookiejar.clear('testserver.local', '/', 'IDP_SESSION')
@@ -672,7 +727,7 @@ def test_passive_auth_middleware_ok(db, app, idp, caplog, settings):
 
     # check passive authentication is tried again
     app.set_cookie('IDP_SESSION', '1')
-    response = app.get('/', headers={'Accept': force_str('text/html')}, status=302)
+    response = app.get('/', headers={'Accept': 'text/html'}, status=302)
     assert urlparse.urlparse(response.location).path == '/login/'
     assert urlparse.parse_qs(urlparse.urlparse(response.location).query, keep_blank_values=True) == {
         'next': ['http://testserver/'],
@@ -680,13 +735,35 @@ def test_passive_auth_middleware_ok(db, app, idp, caplog, settings):
     }
     assert 'MELLON_PASSIVE_TRIED' in app.cookies
 
+    # but not two times
+    response = app.get('/', headers={'Accept': 'text/html'}, status=200)
+    # if session change at IdP
+    app.set_cookie('IDP_SESSION', 'abcd')
+    # then we try again...
+    response = app.get('/', headers={'Accept': 'text/html'}, status=302)
+
+    # ok, let's change again and login
+    app.set_cookie('IDP_SESSION', '5678')
+    response = app.get(reverse('mellon_login') + '?next=/whatever/')
+    url, body, relay_state = idp.process_authn_request_redirect(response['Location'])
+    response = app.post(reverse('mellon_login'), params={'SAMLResponse': body, 'RelayState': relay_state})
+    assert app.session['mellon_opened_session_cookie'] == '5678'
+    assert 'MELLON_PASSIVE_TRIED' not in app.cookies
+    assert '_auth_user_id' in app.session
+    # ok change the idp session id
+    app.set_cookie('IDP_SESSION', '1234')
+    # if we try a request, we are logged out and redirected to try a new passive login
+    response = app.get('/', headers={'Accept': 'text/html'}, status=302)
+    assert '_auth_user_id' not in app.session
+    assert 'MELLON_PASSIVE_TRIED' in app.cookies
+
 
 def test_passive_auth_middleware_no_passive_auth_parameter(db, app, idp, caplog, settings):
     settings.MELLON_OPENED_SESSION_COOKIE_NAME = 'IDP_SESSION'
     assert 'MELLON_PASSIVE_TRIED' not in app.cookies
     # webtest-lint is against unicode
     app.set_cookie('IDP_SESSION', '1')
-    app.get('/?no-passive-auth', headers={'Accept': force_str('text/html')}, status=200)
+    app.get('/?no-passive-auth', headers={'Accept': 'text/html'}, status=200)
 
 
 def test_sso_user_change(db, app, idp, caplog, sp_settings):
@@ -724,6 +801,26 @@ def test_debug_sso(db, app, idp, caplog, sp_settings, settings):
     assert '<saml:Assertion xmlns:saml="urn:oasis:names:tc:SAML:2.0:assertion"' in response_text
     assert '<samlp:Response xmlns:samlp="urn:oasis:names:tc:SAML:2.0:protocol"' in response_text
     assert 'mellon: created new user _' in response_text
+    # check also in logs
+    assert "'email': ['john.doe@gmail.com']" in caplog.text
+    assert '<saml:Assertion xmlns:saml="urn:oasis:names:tc:SAML:2.0:assertion"' in caplog.text
+    assert '<samlp:Response xmlns:samlp="urn:oasis:names:tc:SAML:2.0:protocol"' in caplog.text
+
+
+def test_debug_sso_on_exception(db, app, idp, caplog, sp_settings, settings):
+    settings.DEBUG = True
+    response = app.get(reverse('mellon_debug_login') + '?next=/whatever/')
+    response = response.follow()
+    url, body, relay_state = idp.process_authn_request_redirect(response['Location'])
+
+    def lasso_error(*args, **kwargs):
+        raise lasso.Error
+
+    with mock.patch('lasso.Login.acceptSso', side_effect=lasso_error):
+        response = app.post(reverse('mellon_login'), params={'SAMLResponse': body, 'RelayState': relay_state})
+
+    response_text = unescape(response.text)
+    assert '<saml:Assertion xmlns:saml="urn:oasis:names:tc:SAML:2.0:assertion"' in response_text
 
 
 def test_nonce(db, app, idp, caplog, sp_settings):
@@ -740,3 +837,88 @@ def test_force_authn(db, app, idp, caplog, sp_settings):
 
     response = app.post(reverse('mellon_login'), params={'SAMLResponse': body, 'RelayState': relay_state})
     assert app.session['mellon_session']['force_authn']
+
+
+def test_sso_slo_transient_name_identifier(db, app, idp, caplog, sp_settings):
+    caplog.set_level(logging.WARNING)
+    sp_settings.MELLON_TRANSIENT_FEDERATION_ATTRIBUTE = 'email'
+    response = app.get('/login/')
+    url, body, relay_state = idp.process_authn_request_redirect(
+        response['Location'],
+        name_id={
+            'format': lasso.SAML2_NAME_IDENTIFIER_FORMAT_TRANSIENT,
+            'content': '1234',
+        },
+    )
+    response = app.post('/login/', params={'SAMLResponse': body, 'RelayState': relay_state})
+
+    usi = models.UserSAMLIdentifier.objects.get()
+    assert usi.name_id == 'john.doe@gmail.com'
+    session_index = models.SessionIndex.objects.get(saml_identifier=usi)
+    assert session_index.transient_name_id == '1234'
+
+    response = app.get('/logout/')
+    assert urlparse.urlparse(response['Location']).path == '/singleLogout'
+    url = idp.process_logout_request_redirect(response.location)
+    caplog.clear()
+    response = app.get(url)
+    assert len(caplog.records) == 0, 'logout failed'
+    assert response.location == '/'
+
+
+def test_sso_slo_token(db, app, rf, idp, caplog, django_user_model, freezer):
+    from mellon.views import LogoutView
+
+    caplog.set_level(logging.WARNING)
+    response = app.get('/login/')
+    url, body, relay_state = idp.process_authn_request_redirect(response['Location'])
+    response = app.post('/login/', params={'SAMLResponse': body, 'RelayState': relay_state})
+
+    assert models.SessionIndex.objects.count() == 1
+    assert models.SessionIndex.objects.filter(logout_timestamp__isnull=True).count() == 1
+    request = rf.get('/whatever/')
+    request.session = app.session
+    request.user = django_user_model.objects.get()
+    token_logout_url = LogoutView.make_logout_token_url(request, next_url='/somepath/')
+    assert models.SessionIndex.objects.count() == 1
+    assert models.SessionIndex.objects.filter(logout_timestamp__isnull=False).count() == 1
+    assert token_logout_url
+    app.session.flush()
+    assert '_auth_user_id' not in app.session
+    response = app.get(token_logout_url)
+    assert urlparse.urlparse(response['Location']).path == '/singleLogout'
+    url = idp.process_logout_request_redirect(response.location)
+    caplog.clear()
+    freezer.move_to(datetime.timedelta(minutes=6))
+    response = app.get(url)
+    assert len(caplog.records) == 0, 'logout failed'
+    assert response.location == '/somepath/'
+    assert models.SessionIndex.objects.count() == 0
+
+
+def test_sso_slo_update_of_new_fields(db, app, idp, caplog, sp_settings):
+    response = app.get('/login/')
+    url, body, relay_state = idp.process_authn_request_redirect(response['Location'])
+    response = app.post(
+        reverse('mellon_login'), params={'SAMLResponse': body, 'RelayState': relay_state}
+    ).follow()
+    # violent logout
+    app.session.flush()
+
+    # remove existing fields
+    models.UserSAMLIdentifier.objects.all().update(
+        nid_format=None, nid_name_qualifier=None, nid_sp_name_qualifier=None, nid_sp_provided_id=None
+    )
+
+    response = app.get('/login/')
+    url, body, relay_state = idp.process_authn_request_redirect(response['Location'])
+    response = app.post(
+        reverse('mellon_login'), params={'SAMLResponse': body, 'RelayState': relay_state}
+    ).follow()
+
+    # check logout works
+    response = app.get('/logout/')
+    url = idp.process_logout_request_redirect(response.location)
+    caplog.clear()
+    response = app.get(url)
+    assert len(caplog.records) == 0, 'logout failed'
